﻿--[[
Uber credits must be given to nymbia for the metatables and frames, thanks!
]]
local BS = AceLibrary('Babble-Spell-2.2')
local L = AceLibrary('AceLocale-2.2'):new('ClassTimer')

local sm = AceLibrary('SharedMedia-1.0')

local self = ClassTimer
local ClassTimer = ClassTimer

local GetTime = GetTime
local unpack = unpack
local table_sort = table.sort
local math_ceil = math.ceil


local locked, db = true

local new, del
do
	local cache = setmetatable({}, {__mode='k'})
	function new()
		local t = next(cache)
		if t then
			cache[t] = nil
			return t
		else
			return {}
		end
	end
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		cache[t] = true
		return nil
	end
end
local OnUpdate
do
	local min = L['%dm']
	local function timenum(num)
		if num <= 10 then
			return ('%.1f'):format(num)
		elseif num <= 60 then
			return ('%d'):format(num)
		else
			return min:format(math_ceil(num / 60))
		end
	end
	function OnUpdate(frame)
		local currentTime = GetTime()
		local endTime = frame.endTime
		if currentTime > endTime then
			if locked then locked = nil end
			self:UpdateUnitBars(frame.unit)
		else
			local remaining = (currentTime - frame.startTime)
			frame:SetValue(endTime - remaining)
			frame.timetext:SetText(timenum(endTime - currentTime))
		end
		
	end
end
local function OnShow(frame)
	frame:SetScript('OnUpdate', OnUpdate)
end
local function OnHide(frame)
	frame:SetScript('OnUpdate', nil)
end

local framefactory = {
	__index = function(t,k)
		local bar = CreateFrame('StatusBar', nil, UIParent)
		t[k] = bar
		bar:SetFrameStrata('MEDIUM')
		bar:Hide()
		bar:SetScript('OnShow', OnShow)
		bar:SetScript('OnHide', OnHide)
		bar:SetBackdrop({bgFile = 'Interface\\Tooltips\\UI-Tooltip-Background', tile = true, tileSize = 16})
		bar:SetBackdropColor(0,0,0)
		bar.text = bar:CreateFontString(nil, 'OVERLAY')
		bar.timetext = bar:CreateFontString(nil, 'OVERLAY')
		bar.icon = bar:CreateTexture(nil, 'DIALOG')
		
		self:ApplySettings()
		return bar
	end
}

local bars = {
	target = setmetatable({}, framefactory),
	focus = setmetatable({}, framefactory),
	player = setmetatable({}, framefactory)
}

ClassTimer.Rules = {}
ClassTimer.RuleMenu = {}
ClassTimer.defaults.bars = {
	Units = {	
		['**'] = {
			enable = true,
			buffs = true,
			debuffs = true,
			growup = false,
			showIcons = false,
			icons = true,
			iconSide = L['Left'],
			scale = 1,

			gap = 1,
			spacing = 1,
			offset = 3,
		},
	},
	nametext = true,
	timetext = true,
	
	texture = 'Blizzard',
	width = 120,
	height = 12,
	font = 'Friz Quadrata TT',
	fontsize = 9,
	alpha = 1,
	scale = 1,
	bartext = '%s (%a)',
	
	buffcolor = {0,0.49, 1},
	debuffcolor = {1.0,0.7, 0},
		
	textcolor = {1,1,1},
}

function ClassTimer:RulesInit()
local timerargs = {}
	for k, v in pairs(self:CreateTimers()) do
		timerargs[k] = {
			type = 'group',
			order = 4,
			name = k,
			desc = L['Which Buffs to show.'],
			args = {}
		}
		for _, m in pairs (v) do
			local m = BS[m]
			ClassTimer.defaults.rules[m] = true
			timerargs[k].args[m] = {
				order = 3,
				type = 'toggle',
				name = m,
				desc = L['Enable/Dissable the timer'],
				get = function()
					return self.db.class[m]
				end,
				set = function(u)
					self.db.class[m] = u
				end,
			}
		end
	end
	self.Menu.args.Timers = { type = 'group', order = 1, name = 'Timers', desc = 'Enable or disable timers', args = timerargs, disabled = '~IsActive' }
end

function ClassTimer:RulesSetup()
	self:RegisterBucketEvent('UNIT_AURA', 0.5)
	self:RegisterEvent('PLAYER_TARGET_CHANGED')
	self:RegisterEvent('PLAYER_FOCUS_CHANGED')
	self:RegisterEvent('SharedMedia_SetGlobal', function(mtype, override)
		if mtype == 'statusbar' then
			for k in pairs(bars) do
				for _, v in pairs(k) do
					v:SetStatusBarTexture(sm:Fetch('statusbar', override))
				end
			end
		end
	end)
	ClassTimer:ApplySettings()
end

function ClassTimer:OnDisable()
	locked = nil
	for _, k in pairs(bars) do
		for _, v in pairs(k) do
			v:Hide()
		end
	end
end

function ClassTimer:PLAYER_TARGET_CHANGED()
	self:UpdateUnitBars('target')
end

function ClassTimer:PLAYER_FOCUS_CHANGED()
	self:UpdateUnitBars('focus')
end
function ClassTimer:UNIT_AURA(units)
	for unit in pairs(units) do
		self:UpdateUnitBars(unit)
	end
end

do
	local function sort(a,b)
		if a.isbuff == b.isbuff then
			return a.name < b.name
		else
			return a.isbuff
		end
	end
	local function text(spell, apps)
		local str = self.db.profile.bartext
		str:gsub("%%s", spell)
		if apps and apps > 1 and str:find("%%a") then
			str:gsub("%%a", apps)
		end
		return str
	end
	local tmp = {}
	local called = false -- prevent recursive calls when new bars are created.
	local unitTable = { 'target', 'focus' }
	function ClassTimer:UpdateUnitBars(unit)
	if locked then return end
	--[[for _, k in pairs(unitTable) do
		if k ~= unit and unit ~= 'target' and UnitIsUnit(unit, k) then
			return
		end
	end]]
	if not bars[unit] then return end
		if called then
			return
		end
		called = true
		local db = self.db.profile
		if db.Units[unit].enable then
			local currentTime = GetTime()
			for k in pairs(tmp) do
				tmp[k] = del(tmp[k])
			end
			if db.Units[unit].buffs then
				for i = 1, 32 do
					local name, _, texture, count, duration, remaining = UnitBuff('target', i)
					if not name then
						break
					end
					if duration and duration > 0 and self.db.class[name] then
						local t = new()
						tmp[#tmp+1] = t
						t.name = name
						t.texture = texture
						t.duration = duration
						t.remaining = remaining
						t.count = count
						t.isbuff = true
					end
				end
			end
			if db.Units[unit].debuffs then
				for i = 1, 40 do
					local name, _, texture, count, _, duration, remaining = UnitDebuff('target', i)
					if not name then
						break
					end
					if duration and duration > 0 then
						local t = new()
						tmp[#tmp+1] = t
						t.name = name
						t.texture = texture
						t.duration = duration
						t.remaining = remaining
						t.count = count
					end
				end
			end
			table_sort(tmp, sort)
			local maxindex = 0
			for k,v in ipairs(tmp) do
				maxindex = k
				local bar = bars[unit[k]]
				
				bar.text:SetText(text(v.name, v.count))
				bar.icon:SetTexture(v.texture)
				local elapsed = (v.duration - v.remaining)
				local startTime, endTime = (currentTime - elapsed), (currentTime + v.remaining)
				bar.startTime = startTime
				bar.endTime = endTime
				bar:SetMinMaxValues(startTime, endTime)
				bar:Show()
				if v.isbuff then
					bar:SetStatusBarColor(unpack(db.buffcolor))
				else
					bar:SetStatusBarColor(unpack(db.debuffcolor))
				end
			end
			for i = maxindex+1, #bars[unit] do
				bars[unit][i]:Hide()
			end
		else
			for _, v in ipairs(bars[unit]) do
				v:Hide()
			end
		end
		called = false
	end
end
do
	local function apply(unit, i, bar, db)
		local position, icons, ShowIcons, iconside, gap, spacing, direction
		local bars = bars[unit]
		growup = db.Units[unit].growup
		icons = db.Units[unit].icons
		showIcons = db.Units[unit].showIcons 
		iconSide = db.Units[unit].iconSide
		gap = db.Units[unit].gap
		spacing = db.Units[unit].spacing
		offset = db.Units[unit].offset
		bar:ClearAllPoints()
		bar:SetStatusBarTexture(sm:Fetch('statusbar', db.texture))
		bar:SetWidth(showIcons and 1 or db.width)
		bar:SetHeight(db.height)
		bar:SetScale(db.Units[unit].scale)
		bar:SetAlpha(db.alpha)
		if showIcons then
			bar.text:SetFont(sm:Fetch('font', db.font), db.fontsize)
		end
		if i == 1 then
			direction = growup and 1 or -1 
				if db.Units[unit].x then
					bar:SetPoint('BOTTOMLEFT', UIParent, 'BOTTOMLEFT', db.Units[unit].x, db.Units[unit].y)
				else
					bar:SetPoint('CENTER', UIParent)
				end
		else
			if direction == 1 then
				bar:SetPoint('BOTTOMRIGHT', bars[i-1], 'TOPRIGHT', 0, spacing)
			else 
				bar:SetPoint('TOPRIGHT', bars[i-1], 'BOTTOMRIGHT', 0, -1 * spacing)
			end
		end
		
		local text = bar.text
		if db.nametext and not showIcons then
			text:Show()
			text:ClearAllPoints()
			text:SetFont(sm:Fetch('font', db.font), db.fontsize)
			text:SetShadowColor( 0, 0, 0, 1)
			text:SetShadowOffset( 0.8, -0.8 )
			text:SetTextColor(unpack(db.textcolor))
			text:SetNonSpaceWrap(false)
			text:ClearAllPoints()
			text:SetWidth(db.width)
			text:SetPoint('LEFT', bar, 'LEFT', 2, 0)
			text:SetJustifyH('LEFT')
		else
			text:Hide()
		end

		local timetext = bar.timetext
		if db.timetext then
			timetext:Show()
			timetext:ClearAllPoints()
			timetext:SetFont(sm:Fetch('font', db.font), db.fontsize)
			timetext:SetShadowColor( 0, 0, 0, 1)
			timetext:SetShadowOffset( 0.8, -0.8 )
			timetext:SetTextColor(unpack(db.textcolor))
			timetext:SetNonSpaceWrap(false)
			timetext:ClearAllPoints()
			timetext:SetWidth(db.width)
			timetext:SetPoint(showIcons and 'LEFT' or 'RIGHT', bar, 'RIGHT', -2, 0)
			timetext:SetJustifyH(showIcons and 'LEFT' or 'RIGHT')
		else
			timetext:Hide()
		end
		
		local icon = bar.icon
		if icons then
			icon:Show()
			icon:SetWidth(db.height-1)
			icon:SetHeight(db.height-1)
			icon:SetTexCoord(0.07, 0.93, 0.07, 0.93)
			icon:ClearAllPoints()
			if iconSide == L['Left'] then
				icon:SetPoint('RIGHT', bar, 'LEFT', -1, 0)
			else
				icon:SetPoint('LEFT', bar, 'RIGHT', 1, 0)
			end
		else
			icon:Hide()
		end
		
		return direction
	end
	function ClassTimer:ApplySettings()
		local db = ClassTimer.db.profile
		for i, v in pairs(bars.target) do
			apply('target', i, v, db)
		end
		for i, v in pairs(bars.focus) do
			apply('focus', i, v, db)
		end
	
	end
end

do
	local function dragstart()
		this:StartMoving()
	end
	local function dragstop()
		self.db.profile.Units[this.unit].x = this:GetLeft()
		self.db.profile.Units[this.unit].y = this:GetBottom()
		this:StopMovingOrSizing()
	end
	local function set(table, value)
		local s1, s2 = unpack(table)
		if s2 then
			ClassTimer.db.profile.Units[s1][s2] = value
		else
			ClassTimer.db.profile[s1] = value
		end
			ClassTimer.ApplySettings()
	end
	local function get(table)
		local s1, s2 = unpack(table)
		if s2 then
			return ClassTimer.db.profile.Units[s1][s2]
		else
			return ClassTimer.db.profile[s1]
		end
	end
	local function setcolor(field, ...)
		ClassTimer.db.profile[field] = {...}
		ClassTimer.ApplySettings()
	end
	local function getcolor(field)
		return unpack(ClassTimer.db.profile[field])
	end
	local function hidefocusoptions()	
		return not ClassTimer.db.profile.Units.focus.enable
	end
	local function hidetargetoptions()
		return not ClassTimer.db.profile.Units.target.enable
	end
	local function Unit(unit)
		return {
			type = 'group',
			name = L[unit],
			desc = L[unit],
			order = 101,
			args = {
				show = {
					type = 'toggle',
					name = L['Enable %s']:format(L[unit]),
					desc = L['Show buffs/debuffs for your %s']:format(L[unit]),
					get = get,
					set = set,
					passValue = {unit, 'enable'},
					order = 99,
				},
				lock = {
					type = 'toggle',
					name = L['Lock'],
					desc = L['Toggle Cast Bar lock'],
					order = 100,
					get = function()
						return locked
					end,
					set = function(v)
						if v then
							bars[unit][1]:EnableMouse(false)
							bars[unit][1]:SetMovable(false)
							bars[unit][1]:SetScript('OnDragStart', nil)
							bars[unit][1]:SetScript('OnDragStop', nil)
							ClassTimer:UpdateUnitBars(unit)
							bars[unit].unit = nil
						else
							ClassTimer:OnDisable()
							local bar = bars[unit][1]
							bar:Show()
							bar.icon:SetTexture(self.hasIcon)
							bar.text:SetText('Drag to move')
							bar.startTime = GetTime()
							bar.endTime = GetTime() + 120
							bar:SetMinMaxValues(GetTime(),  GetTime() + 120)
							bar:SetStatusBarColor(unpack(self.db.profile.buffcolor))
							bar:RegisterForDrag('LeftButton')
							bar.unit = unit
							bar:EnableMouse(true)
							bar:SetMovable(true)
							bar:SetScript('OnDragStart', dragstart)
							bar:SetScript('OnDragStop', dragstop)
							bar:SetAlpha(1)
						end
						locked = v
					end,
					order = 100,
				},
				buffs = {
					type = 'toggle',
					name = L['Enable Buffs'],
					desc = L['Show buffs for your %s']:format(L[unit]),
					get = get,
					set = set,
					passValue = {unit, 'buffs'},
					hidden = hidefocusoptions,
				},
				debuffs = {
					type = 'toggle',
					name = L['Enable Debuffs'],
					desc = L['Show debuffs for your %s']:format(L[unit]),
					get = get,
					set = set,
					passValue = {unit, 'debuffs'},
					hidden = hidefocusoptions,
				},
				growUp = {
					type = 'toggle',
					name = L['Grow Up'],
					desc = L['Set %s bars to grow up']:format(L[unit]),
					get = get,
					set = set,
					passValue = {unit, 'position'},
					hidden = hidefocusoptions,
				},
				gap = {
					type = 'range',
					name = L['Gap'],
					desc = L['Tweak the vertical position of the bars for your %s']:format(L[unit]),
					min = -10,
					max = 10,
					step = 1,
					get = get,
					set = set,
					passValue = {unit, 'gap'},
					order = 101,
					hidden = hidefocusoptions,
				},
				spacing = {
					type = 'range',
					name = L['Spacing'],
					desc = L['Tweak the space between bars for your %s']:format(L[unit]),
					min = -5,
					max = 5,
					step = 1,
					get = get,
					set = set,
					passValue = {unit, 'spacing'},
					order = 101,
					hidden = hidefocusoptions,
				},
				scale = {
					type = 'range',
					name = L['Scale'],
					desc = L['Set the scale of the bars for your %s']:format(L[unit]),
					get = get,
					set = set,
					min = .1,
					max = 2,
					step = 1,
					passValue = {unit, 'scale'},
				},
				offset = {
					type = 'range',
					name = L['Offset'],
					desc = L['Tweak the horizontal position of the bars for your %s']:format(L[unit]),
					min = -10,
					max = 10,
					step = 1,
					get = get,
					set = set,
					passValue = {unit, 'offset'},
					order = 101,
					hidden = hidefocusoptions,
				},
				showicons = {
					type = 'toggle',
					name = L['Show Icons'],
					desc = L['Show icons on buffs and debuffs for your %s']:format(L[unit]),
					get = get,
					set = set,
					passValue = {unit, 'icons'},
					order = 103,
					hidden = hidefocusoptions,
				},
				showiconsOnly = {
					type = 'toggle',
					name = L['Show Only Icons'],
					desc = L['Show only icons and timeleft for your %s']:format(L[unit]),
					get = get,
					set = set,
					passValue = {unit, 'showIcons'},
					order = 103,
					hidden = hidefocusoptions,
				},
				iconside = {
					type = 'text',
					name = L['Icon Position'],
					desc = L['Set the side of the bar that the icon appears on'],
					get = get,
					set = set,
					validate = {L['Left'], L['Right']},
					passValue = {unit, 'iconSide'},
					order = 104,
					hidden = hidefocusoptions,
				},
			}
		}
	end

	ClassTimer.Menu.args.BarSettings = {
		type = 'group',
		name = L['Bar Settings'],
		desc = L['Bar Settings'],
		order = 602,
		disabled = '~IsActive',
		args = {
			focus = Unit('focus'),
			target = Unit('target'),
			player = Unit('player'),
			texture = {
				type = 'text',
				name = L['Texture'],
				desc = L['Set the bar Texture'],
				validate = sm:List('statusbar'),
				order = 103,
				get = get,
				set = set,
				passValue = {'texture'},
			},
			width = {
				type = 'range',
				name = L['Bar Width'],
				desc = L['Set the width of the bars'],
				min = 50,
				max = 300,
				step = 1,
				get = get,
				set = set,
				passValue = {'width'},
				order = 104,
			},
			height = {
				type = 'range',
				name = L['Bar Height'],
				desc = L['Set the height of the bars'],
				min = 4,
				max = 25,
				step = 1,
				get = get,
				set = set,
				passValue = {'height'},
				order = 105,
			},
			nametext = {
				type = 'toggle',
				name = L['Name Text'],
				desc = L['Display the names of buffs/debuffs on their bars'],
				get = get,
				set = set,
				passValue = {'nametext'},
				order = 106,
			},
			timetext = {
				type = 'toggle',
				name = L['Time Text'],
				desc = L['Display the time remaining on buffs/debuffs on their bars'],
				get = get,
				set = set,
				passValue = {'timetext'},
				order = 107,
			},
			font = {
				type = 'text',
				name = L['Font'],
				desc = L['Set the font used in the bars'],
				validate = sm:List('font'),
				get = get,
				set = set,
				passValue = {'font'},
				order = 108,
			},
			textcolor = {
				type = 'color',
				name = L['Text Color'],
				desc = L['Set the color of the text for the bars'],
				get = getcolor,
				set = setcolor,
				passValue = 'textcolor',
				order = 109,
			},
			fontsize = {
				type = 'range',
				name = L['Font Size'],
				desc = L['Set the font size for the bars'],
				min = 3,
				max = 15,
				step = 1,
				get = get,
				set = set,
				passValue = {'fontsize'},
				order = 110,
			},
			alpha = {
				type = 'range',
				name = L['Alpha'],
				desc = L['Set the alpha of the bars'],
				min = 0.05,
				max = 1,
				step = 0.05,
				isPercent = true,
				get = get,
				set = set,
				passValue = {'alpha'},
				order = 111,
			},
			color = {
				type = 'color',
				name = L['Color'],
				desc = L['Set the color of the bars for buffs'],
				get = getcolor,
				set = setcolor,
				passValue = 'buffcolor',
				order = 129,
			},
			debuffcolor = {
				type = 'color',
				name = L['Debuff Color'],
				desc = L['Set the color of the bars for debuffs'],
				get = getcolor,
				set = setcolor,
				passValue = 'debuffcolor',
				order = 130,
			},
			text = {
				type = 'text',
				name = L['Bar Text'],
				desc = L['Set the bar text'],
				get = get,
				set = set,
				usage = L['<%s for spell, %a for applications. %s must be included>'],
				validate = function(name) return string.find(name, "%%s") end,
				passValue = {'bartext'},
			},
		},
	}
end
